@@ -36,9 +36,30 @@
 #include "main.h"
 #include "extern.h"
 
-/*----------------------------------------------------------------------
- *                       Completion functions
- *----------------------------------------------------------------------*/
+
+/*
+ * Structure
+ */
+struct Completion
+{
+#define FIELD(ty, name) ty name;
+#define FIELD_STR(name) const char *name;
+#include "completion.h"
+#undef FIELD
+#undef FIELD_STR
+};
+
+#define FIELD(ty, field)                        \
+  GETTER (Completion, completion, ty, field)    \
+  SETTER (Completion, completion, ty, field)
+
+#define FIELD_STR(field)                                        \
+  GETTER (Completion, completion, const char *, field)          \
+  STR_SETTER (Completion, completion, field)
+
+#include "completion.h"
+#undef FIELD
+#undef FIELD_STR
 
 /*
  * List methods for completions and matches
@@ -99,11 +120,13 @@ void
 completion_scroll_up (void)
 {
   Window *wp, *old_wp = cur_wp;
+  Point pt;
 
   wp = find_window ("*Completions*");
   assert (wp != NULL);
   set_current_window (wp);
-  if (cur_bp->pt.n == cur_bp->last_line || !FUNCALL (scroll_up))
+  pt = get_buffer_pt (cur_bp);
+  if (pt.n == get_buffer_last_line (cur_bp) || !FUNCALL (scroll_up))
     gotobob ();
   set_current_window (old_wp);
 
@@ -117,11 +140,13 @@ void
 completion_scroll_down (void)
 {
   Window *wp, *old_wp = cur_wp;
+  Point pt;
 
   wp = find_window ("*Completions*");
   assert (wp != NULL);
   set_current_window (wp);
-  if (cur_bp->pt.n == 0 || !FUNCALL (scroll_down))
+  pt = get_buffer_pt (cur_bp);
+  if (pt.n == 0 || !FUNCALL (scroll_down))
     {
       gotoeob ();
       resync_redisplay ();
@@ -157,7 +182,7 @@ completion_print (gl_list_t l, size_t size)
   size_t i, j, col, max, numcols;
 
   max = calculate_max_length (l, size) + 5;
-  numcols = (cur_wp->ewidth - 1) / max;
+  numcols = (get_window_ewidth (cur_wp) - 1) / max;
 
   bprintf ("Possible completions are:\n");
   for (i = col = 0; i < MIN (size, gl_list_size (l)); i++)
@@ -195,7 +220,7 @@ static void
 popup_completion (Completion * cp, int allflag, size_t num)
 {
   cp->flags |= CFLAG_POPPEDUP;
-  if (head_wp->next == NULL)
+  if (get_window_next (head_wp) == NULL)
     cp->flags |= CFLAG_CLOSE;
 
   write_temp_buffer ("*Completions*", true, write_completion, cp, allflag, num);
@@ -234,12 +259,12 @@ completion_readdir (Completion * cp, astr as)
      in which case it's considered to be entirely dirname */
   s1 = xstrdup (astr_cstr (as));
   s2 = xstrdup (astr_cstr (as));
-  if (*astr_char (as, astr_len (as) - 1) != '/')
+  if (astr_get (as, astr_len (as) - 1) != '/')
     {
       pdir = dirname (s1);
       /* Append `/' to pdir */
       astr_cat_cstr (bs, pdir);
-      if (*astr_char (bs, astr_len (bs) - 1) != '/')
+      if (astr_get (bs, astr_len (bs) - 1) != '/')
         astr_cat_char (bs, '/');
       pdir = astr_cstr (bs);
       base = base_name (s2);
