@@ -54,13 +54,7 @@ is_regular_file (const char *filename)
 {
   struct stat st;
 
-  if (stat (filename, &st) == -1)
-    {
-      if (errno == ENOENT)
-        return true;
-      return false;
-    }
-  if (S_ISREG (st.st_mode))
+  if (stat (filename, &st) == 0 && S_ISREG (st.st_mode))
     return true;
 
   return false;
@@ -69,7 +63,7 @@ is_regular_file (const char *filename)
 /*
  * Return the current directory, if available.
  */
-static astr
+astr
 agetcwd (void)
 {
   char *s = getcwd (NULL, 0);
@@ -101,7 +95,7 @@ expand_path (astr path)
       astr_cat (epath, cwd);
       astr_delete (cwd);
       if (astr_len (epath) == 0 ||
-          *astr_char (epath, astr_len (epath) - 1) != '/')
+          astr_get (epath, astr_len (epath) - 1) != '/')
         astr_cat_char (epath, '/');
     }
 
@@ -116,7 +110,7 @@ expand_path (astr path)
               astr_truncate (epath, 0);
             }
           if (astr_len (epath) == 0 ||
-              *astr_char (epath, astr_len (epath) - 1) != '/')
+              astr_get (epath, astr_len (epath) - 1) != '/')
             astr_cat_char (epath, '/');
         }
       else if (*p == '~' && (p == sp || p[-1] == '/'))
@@ -158,9 +152,9 @@ expand_path (astr path)
         }
       else if (*p == '.' && p[1] == '.' && (p[2] == '/' || p[2] == '\0'))
         { /* Got `..'. */
-          if (astr_len (epath) >= 1 && *astr_char (epath, astr_len (epath) - 1) == '/')
+          if (astr_len (epath) >= 1 && astr_get (epath, astr_len (epath) - 1) == '/')
             astr_truncate (epath, astr_len (epath) - 1);
-          while (*astr_char (epath, astr_len (epath) - 1) != '/' && astr_len (epath) >= 1)
+          while (astr_get (epath, astr_len (epath) - 1) != '/' && astr_len (epath) >= 1)
             astr_truncate (epath, astr_len (epath) - 1);
           p += 2;
         }
@@ -188,14 +182,15 @@ compact_path (astr path)
   if (pw != NULL)
     {
       /* Replace `/userhome/' (if found) with `~/'. */
-      size_t i = strlen (pw->pw_dir);
-      if (!strncmp (pw->pw_dir, astr_cstr (path), i))
+      size_t homelen = strlen (pw->pw_dir);
+      if (!strncmp (pw->pw_dir, astr_cstr (path),
+                    MIN (homelen, astr_len (path))))
         {
           astr buf = astr_new_cstr ("~/");
           if (!strcmp (pw->pw_dir, "/"))
-            astr_cat_cstr (buf, astr_char (path, 1));
+            astr_cat_cstr (buf, astr_cstr (path) + 1);
           else
-            astr_cat_cstr (buf, astr_char (path, i + 1));
+            astr_cat_cstr (buf, astr_cstr (path) + homelen + 1);
           astr_cpy (path, buf);
           astr_delete (buf);
         }
@@ -213,14 +208,14 @@ get_buffer_dir (void)
   astr buf;
   const char *p, *q;
 
-  if (cur_bp->filename != NULL)
+  if (get_buffer_filename (cur_bp) != NULL)
     /* If the current buffer has a filename, get the current directory
        name from it. */
-    buf = astr_new_cstr (cur_bp->filename);
+    buf = astr_new_cstr (get_buffer_filename (cur_bp));
   else
     { /* Get the current directory name from the system. */
       buf = agetcwd ();
-      if (astr_len (buf) != 0 && *astr_char (buf, astr_len (buf) - 1) != '/')
+      if (astr_len (buf) != 0 && astr_get (buf, astr_len (buf) - 1) != '/')
         astr_cat_char (buf, '/');
     }
 
@@ -266,26 +261,29 @@ static void
 read_from_disk (const char *filename)
 {
   Line *lp;
-  int i, size, first_eol = true;
+  int i, size;
+  bool first_eol = true;
   char *this_eol_type;
   size_t eol_len = 0, total_eols = 0;
   char buf[BUFSIZ];
   FILE *fp = fopen (filename, "r");
+  Point pt;
 
   if (fp == NULL)
     {
       if (errno != ENOENT)
         {
           minibuf_write ("%s: %s", filename, strerror (errno));
-          cur_bp->flags |= BFLAG_READONLY;
+          set_buffer_readonly (cur_bp, true);
         }
       return;
     }
 
   if (!check_writable (filename))
-    cur_bp->flags |= BFLAG_READONLY;
+    set_buffer_readonly (cur_bp, true);
 
-  lp = cur_bp->pt.p;
+  pt = get_buffer_pt (cur_bp);
+  lp = pt.p;
 
   /* Read first chunk and determine EOL type. */
   size = fread (buf, 1, BUFSIZ, fp);
@@ -309,31 +307,32 @@ read_from_disk (const char *filename)
               if (first_eol)
                 {
                   /* This is the first end-of-line. */
-                  cur_bp->eol = this_eol_type;
+                  set_buffer_eol (cur_bp, this_eol_type);
                   first_eol = false;
                 }
-              else if (cur_bp->eol != this_eol_type)
+              else if (get_buffer_eol (cur_bp) != this_eol_type)
                 {
                   /* This EOL is different from the last; arbitrarily choose
                      LF. */
-                  cur_bp->eol = coding_eol_lf;
+                  set_buffer_eol (cur_bp, coding_eol_lf);
                   break;
                 }
             }
         }
 
       /* Process this and subsequent chunks into lines. */
-      eol_len = strlen (cur_bp->eol);
+      eol_len = strlen (get_buffer_eol (cur_bp));
       do
         {
           for (i = 0; i < size; i++)
             {
-              if (strncmp (cur_bp->eol, buf + i, eol_len) != 0)
-                astr_cat_char (lp->text, buf[i]);
+              if (strncmp (get_buffer_eol (cur_bp), buf + i, eol_len) != 0)
+                astr_cat_char (get_line_text (lp), buf[i]);
               else
                 {
                   lp = line_insert (lp, astr_new ());
-                  ++cur_bp->last_line;
+                  set_buffer_last_line (cur_bp,
+                                        get_buffer_last_line (cur_bp) + 1);
                   i += eol_len - 1;
                 }
             }
@@ -341,9 +340,11 @@ read_from_disk (const char *filename)
       while ((size = fread (buf, 1, BUFSIZ, fp)) > 0);
     }
 
-  lp->next = cur_bp->lines;
-  cur_bp->lines->prev = lp;
-  cur_bp->pt.p = cur_bp->lines->next;
+  set_line_next (lp, get_buffer_lines (cur_bp));
+  set_line_prev (get_buffer_lines (cur_bp), lp);
+  pt = get_buffer_pt (cur_bp);
+  pt.p = get_line_next (get_buffer_lines (cur_bp));
+  set_buffer_pt (cur_bp, pt);
 
   fclose (fp);
 }
@@ -352,32 +353,23 @@ bool
 find_file (const char *filename)
 {
   Buffer *bp;
-  char *s;
 
-  for (bp = head_bp; bp != NULL; bp = bp->next)
-    if (bp->filename != NULL && !strcmp (bp->filename, filename))
+  for (bp = head_bp; bp != NULL; bp = get_buffer_next (bp))
+    if (get_buffer_filename (bp) != NULL &&
+        !strcmp (get_buffer_filename (bp), filename))
       {
         switch_to_buffer (bp);
         return true;
       }
 
-  s = make_buffer_name (filename);
-  if (strlen (s) < 1)
+  if (exist_file (filename) && !is_regular_file (filename))
     {
-      free (s);
+      minibuf_error ("File exists but could not be read");
       return false;
     }
 
-  if (!is_regular_file (filename))
-    {
-      minibuf_error ("%s is not a regular file", filename);
-      waitkey (WAITKEY_DEFAULT);
-      return false;
-    }
-
-  bp = create_buffer (s);
-  free (s);
-  bp->filename = xstrdup (filename);
+  bp = buffer_new ();
+  set_buffer_names (bp, filename);
 
   switch_to_buffer (bp);
   read_from_disk (filename);
@@ -387,20 +379,6 @@ find_file (const char *filename)
   return true;
 }
 
-static Completion *
-make_buffer_completion (void)
-{
-  Buffer *bp;
-  Completion *cp;
-
-  cp = completion_new (false);
-  for (bp = head_bp; bp != NULL; bp = bp->next)
-    gl_sortedlist_add (cp->completions, completion_strcmp,
-                       xstrdup (bp->name));
-
-  return cp;
-}
-
 DEFUN ("find-file", find_file)
 /*+
 Edit the specified file.
@@ -432,36 +410,10 @@ Use @kbd{M-x toggle-read-only} to permit editing.
 {
   ok = FUNCALL (find_file);
   if (ok == leT)
-    cur_bp->flags |= BFLAG_READONLY;
+    set_buffer_readonly (cur_bp, true);
 }
 END_DEFUN
 
-/*
- * Check if the buffer has been modified.  If so, asks the user if
- * he/she wants to save the changes.  If the response is positive, return
- * true, else false.
- */
-static int
-check_modified_buffer (Buffer * bp)
-{
-  if (bp->flags & BFLAG_MODIFIED && !(bp->flags & BFLAG_NOSAVE))
-    for (;;)
-      {
-        int ans = minibuf_read_yesno
-          ("Buffer %s modified; kill anyway? (yes or no) ", bp->name);
-        if (ans == -1)
-          {
-            FUNCALL (keyboard_quit);
-            return false;
-          }
-        else if (!ans)
-          return false;
-        break;
-      }
-
-  return true;
-}
-
 DEFUN ("find-alternate-file", find_alternate_file)
 /*+
 Find the file specified by the user, select its buffer, kill previous buffer.
@@ -469,8 +421,8 @@ If the current buffer now contains an empty file that you just visited
 (presumably by mistake), use this command to visit the file you really want.
 +*/
 {
-  const char *buf = cur_bp->filename;
-  char *base = base_name (cur_bp->filename);
+  const char *buf = get_buffer_filename (cur_bp);
+  char *base = base_name (buf);
   char *ms = minibuf_read_filename ("Find alternate: ", buf, base);
 
   ok = leNIL;
@@ -491,17 +443,17 @@ END_DEFUN
 DEFUN_ARGS ("switch-to-buffer", switch_to_buffer,
             STR_ARG (buffer))
 /*+
-Select to the user specified buffer in the current window.
+Select buffer @i{buffer} in the current window.
 +*/
 {
-  Buffer *bp = ((cur_bp->next != NULL) ? cur_bp->next : head_bp);
+  Buffer *bp = ((get_buffer_next (cur_bp) != NULL) ? get_buffer_next (cur_bp) : head_bp);
 
   STR_INIT (buffer)
   else
     {
       Completion *cp = make_buffer_completion ();
       buffer = minibuf_read_completion ("Switch to buffer (default %s): ",
-                                        "", cp, NULL, bp->name);
+                                        "", cp, NULL, get_buffer_name (bp));
       free_completion (cp);
 
       if (buffer == NULL)
@@ -512,11 +464,17 @@ Select to the user specified buffer in the current window.
     {
       if (buffer && buffer[0] != '\0')
         {
-          bp = find_buffer (buffer, false);
+          bp = find_buffer (buffer);
           if (bp == NULL)
             {
-              bp = find_buffer (buffer, true);
-              bp->flags = BFLAG_NEEDNAME | BFLAG_NOSAVE;
+              bp = find_buffer (buffer);
+              if (bp == NULL)
+                {
+                  bp = buffer_new ();
+                  set_buffer_name (bp, buffer);
+                }
+              set_buffer_needname (bp, true);
+              set_buffer_nosave (bp, true);
             }
         }
 
@@ -527,135 +485,12 @@ Select to the user specified buffer in the current window.
 }
 END_DEFUN
 
-/*
- * Remove the specified buffer from the buffer list and deallocate
- * its space.  Avoid killing the sole buffers and creates the scratch
- * buffer when required.
- */
-void
-kill_buffer (Buffer * kill_bp)
-{
-  Buffer *next_bp;
-
-  if (kill_bp->next != NULL)
-    next_bp = kill_bp->next;
-  else
-    next_bp = head_bp;
-
-  if (next_bp == kill_bp)
-    {
-      Window *wp;
-      Buffer *new_bp = create_buffer (cur_bp->name);
-      /* If this is the sole buffer available, then remove the contents
-         and set the name to `*scratch*' if it is not already set. */
-      assert (cur_bp == kill_bp);
-
-      free_buffer (cur_bp);
-
-      /* Scan all the windows that display this buffer. */
-      for (wp = head_wp; wp != NULL; wp = wp->next)
-        if (wp->bp == cur_bp)
-          {
-            wp->bp = new_bp;
-            wp->topdelta = 0;
-            wp->saved_pt = NULL;	/* It was freed. */
-          }
-
-      head_bp = cur_bp = new_bp;
-      head_bp->next = NULL;
-      cur_bp->flags = BFLAG_NOSAVE | BFLAG_NEEDNAME | BFLAG_TEMPORARY;
-      if (strcmp (cur_bp->name, "*scratch*") != 0)
-        {
-          set_buffer_name (cur_bp, "*scratch*");
-          if (cur_bp->filename != NULL)
-            {
-              free (cur_bp->filename);
-              cur_bp->filename = NULL;
-            }
-        }
-    }
-  else
-    {
-      Buffer *bp;
-      Window *wp;
-
-      /* Search for windows displaying the buffer to kill. */
-      for (wp = head_wp; wp != NULL; wp = wp->next)
-        if (wp->bp == kill_bp)
-          {
-            wp->bp = next_bp;
-            wp->topdelta = 0;
-            wp->saved_pt = NULL;	/* The marker will be freed. */
-          }
-
-      /* Remove the buffer from the buffer list. */
-      cur_bp = next_bp;
-      if (head_bp == kill_bp)
-        head_bp = head_bp->next;
-      for (bp = head_bp; bp->next != NULL; bp = bp->next)
-        if (bp->next == kill_bp)
-          {
-            bp->next = bp->next->next;
-            break;
-          }
-
-      free_buffer (kill_bp);
-
-      thisflag |= FLAG_NEED_RESYNC;
-    }
-}
-
-DEFUN_ARGS ("kill-buffer", kill_buffer,
-            STR_ARG (buffer))
-/*+
-Kill buffer BUFFER.
-With a nil argument, kill the current buffer.
-+*/
-{
-  Buffer *bp;
-
-  STR_INIT (buffer)
-  else
-    {
-      Completion *cp = make_buffer_completion ();
-      buffer = minibuf_read_completion ("Kill buffer (default %s): ",
-                                        "", cp, NULL, cur_bp->name);
-      free_completion (cp);
-      if (buffer == NULL)
-        ok = FUNCALL (keyboard_quit);
-    }
-
-  if (buffer && buffer[0] != '\0')
-    {
-      bp = find_buffer (buffer, false);
-      if (bp == NULL)
-        {
-          minibuf_error ("Buffer `%s' not found", buffer);
-          free ((char *) buffer);
-          ok = leNIL;
-        }
-    }
-  else
-    bp = cur_bp;
-
-  if (ok == leT)
-    {
-      if (!check_modified_buffer (bp))
-        ok = leNIL;
-      else
-        kill_buffer (bp);
-    }
-
-  STR_FREE (buffer);
-}
-END_DEFUN
-
 static size_t
 insert_lines (size_t n, size_t end, size_t last, Line *from_lp)
 {
-  for (; n < end; n++, from_lp = from_lp->next)
+  for (; n < end; n++, from_lp = get_line_next (from_lp))
     {
-      insert_astr (from_lp->text);
+      insert_astr (get_line_text (from_lp));
       if (n < last)
         insert_newline ();
     }
@@ -665,18 +500,19 @@ insert_lines (size_t n, size_t end, size_t last, Line *from_lp)
 static void
 insert_buffer (Buffer * bp)
 {
-  Line *old_next = bp->pt.p->next;
-  astr old_cur_line = astr_cpy (astr_new (), bp->pt.p->text);
-  size_t old_cur_n = bp->pt.n, old_lines = bp->last_line;
+  Line *old_next = get_line_next (get_buffer_pt (bp).p);
+  astr old_cur_line = astr_cpy (astr_new (), get_line_text (get_buffer_pt (bp).p));
+  size_t old_cur_n = get_buffer_pt (bp).n, old_lines = get_buffer_last_line (bp);
   size_t size = calculate_buffer_size (bp);
 
-  undo_save (UNDO_REPLACE_BLOCK, cur_bp->pt, 0, size);
+  undo_save (UNDO_REPLACE_BLOCK, get_buffer_pt (cur_bp), 0, size);
   undo_nosave = true;
-  insert_lines (0, old_cur_n, old_lines, bp->lines->next);
+  insert_lines (0, old_cur_n, old_lines, get_line_next (get_buffer_lines (bp)));
   insert_astr (old_cur_line);
   if (old_cur_n < old_lines)
     insert_newline ();
   insert_lines (old_cur_n + 1, old_lines, old_lines, old_next);
+  astr_delete (old_cur_line);
   undo_nosave = false;
 }
 
@@ -687,7 +523,7 @@ Insert after point the contents of BUFFER.
 Puts mark after the inserted text.
 +*/
 {
-  Buffer *def_bp = ((cur_bp->next != NULL) ? cur_bp->next : head_bp);
+  Buffer *def_bp = ((get_buffer_next (cur_bp) != NULL) ? get_buffer_next (cur_bp) : head_bp);
   Completion *cp;
 
   if (warn_if_readonly_buffer ())
@@ -698,7 +534,7 @@ Puts mark after the inserted text.
     {
       cp = make_buffer_completion ();
       buffer = minibuf_read_completion ("Insert buffer (default %s): ",
-                                        "", cp, NULL, def_bp->name);
+                                        "", cp, NULL, get_buffer_name (def_bp));
       if (buffer == NULL)
         ok = FUNCALL (keyboard_quit);
       free_completion (cp);
@@ -710,7 +546,7 @@ Puts mark after the inserted text.
 
       if (buffer && buffer[0] != '\0')
         {
-          bp = find_buffer (buffer, false);
+          bp = find_buffer (buffer);
           if (bp == NULL)
             {
               minibuf_error ("Buffer `%s' not found", buffer);
@@ -756,7 +592,7 @@ insert_file (const char *filename)
     }
 
   lseek (fd, 0, SEEK_SET);
-  undo_save (UNDO_REPLACE_BLOCK, cur_bp->pt, 0, size);
+  undo_save (UNDO_REPLACE_BLOCK, get_buffer_pt (cur_bp), 0, size);
   undo_nosave = true;
   while ((size = read (fd, buf, BUFSIZ)) > 0)
     insert_nstring (buf, size);
@@ -889,7 +725,7 @@ copy_file (const char *source, const char *dest)
 static int
 raw_write_to_disk (Buffer * bp, const char *filename, mode_t mode)
 {
-  ssize_t eol_len = (ssize_t) strlen (bp->eol), written;
+  ssize_t eol_len = (ssize_t) strlen (get_buffer_eol (bp)), written;
   Line *lp;
   int ret = 0;
   int fd = creat (filename, mode);
@@ -898,19 +734,21 @@ raw_write_to_disk (Buffer * bp, const char *filename, mode_t mode)
     return -1;
 
   /* Save the lines. */
-  for (lp = bp->lines->next; lp != bp->lines; lp = lp->next)
+  for (lp = get_line_next (get_buffer_lines (bp));
+       lp != get_buffer_lines (bp);
+       lp = get_line_next (lp))
     {
-      ssize_t len = (ssize_t) astr_len (lp->text);
+      ssize_t len = (ssize_t) astr_len (get_line_text (lp));
 
-      written = write (fd, astr_cstr (lp->text), len);
+      written = write (fd, astr_cstr (get_line_text (lp)), len);
       if (written != len)
         {
           ret = written;
           break;
         }
-      if (lp->next != bp->lines)
+      if (get_line_next (lp) != get_buffer_lines (bp))
         {
-          written = write (fd, bp->eol, eol_len);
+          written = write (fd, get_buffer_eol (bp), eol_len);
           if (written != eol_len)
             {
               ret = written;
@@ -939,7 +777,7 @@ create_backup_filename (const char *filename, const char *backupdir)
       astr buf = astr_new ();
 
       astr_cpy_cstr (buf, backupdir);
-      if (*astr_char (buf, astr_len (buf) - 1) != '/')
+      if (astr_get (buf, astr_len (buf) - 1) != '/')
         astr_cat_char (buf, '/');
       while (*filename)
         {
@@ -964,27 +802,25 @@ create_backup_filename (const char *filename, const char *backupdir)
 }
 
 /*
- * Write the buffer contents to a file.  Create a backup file if specified
- * by the user variables.
+ * Write the buffer contents to a file.
+ * Create a backup file if specified by the user variables.
  */
 static int
-write_to_disk (Buffer * bp, char *filename)
+write_to_disk (Buffer * bp, const char *filename)
 {
   int fd, backup = get_variable_bool ("make-backup-files"), ret;
   const char *backupdir = get_variable_bool ("backup-directory") ?
     get_variable ("backup-directory") : NULL;
 
-  /*
-   * Make backup of original file.
-   */
-  if (!(bp->flags & BFLAG_BACKUP) && backup
+  /* Make backup of original file. */
+  if (!get_buffer_backup (bp) && backup
       && (fd = open (filename, O_RDWR, 0)) != -1)
     {
       astr bfilename;
       close (fd);
       bfilename = create_backup_filename (filename, backupdir);
       if (bfilename && copy_file (filename, astr_cstr (bfilename)))
-        bp->flags |= BFLAG_BACKUP;
+        set_buffer_backup (bp, true);
       else
         {
           minibuf_error ("Cannot make backup file: %s", strerror (errno));
@@ -1007,52 +843,71 @@ write_to_disk (Buffer * bp, char *filename)
   return true;
 }
 
-static int
-save_buffer (Buffer * bp)
+static le *
+write_buffer (Buffer *bp, bool needname, bool confirm,
+              const char *name, const char *prompt)
 {
-  char *ms, *fname = bp->filename != NULL ? bp->filename : bp->name;
-  int ms_is_from_minibuffer = 0;
+  bool ans = true;
+  bool name_from_minibuffer = false;
+  le * ok = leT;
 
-  if (!(bp->flags & BFLAG_MODIFIED))
-    minibuf_write ("(No changes need to be saved)");
-  else
+  if (needname)
     {
-      if (bp->flags & BFLAG_NEEDNAME)
+      name = minibuf_read_filename (prompt, "", NULL);
+      name_from_minibuffer = true;
+      if (name == NULL)
+        return FUNCALL (keyboard_quit);
+      if (name[0] == '\0')
         {
-          ms = minibuf_read_filename ("File to save in: ", fname, NULL);
-          if (ms == NULL)
-            {
-              FUNCALL (keyboard_quit);
-              return false;
-            }
-          ms_is_from_minibuffer = 1;
-          if (ms[0] == '\0')
-            {
-              free (ms);
-              return false;
-            }
+          free ((char *) name);
+          return leNIL;
+        }
+      confirm = true;
+    }
 
-          set_buffer_filename (bp, ms);
+  if (confirm && exist_file (name))
+    {
+      ans = minibuf_read_yn ("File `%s' exists; overwrite? (y or n) ", name);
+      if (ans == -1)
+        FUNCALL (keyboard_quit);
+      if (ans == false)
+        minibuf_error ("Canceled");
+      if (ans != true)
+        ok = leNIL;
+    }
 
-          bp->flags &= ~BFLAG_NEEDNAME;
+  if (ans == true)
+    {
+      if (name != get_buffer_filename (bp))
+        set_buffer_names (bp, name);
+      set_buffer_needname (bp, false);
+      set_buffer_temporary (bp, false);
+      set_buffer_nosave (bp, false);
+      if (write_to_disk (bp, name))
+        {
+          minibuf_write ("Wrote %s", name);
+          set_buffer_modified (bp, false);
+          undo_set_unchanged (get_buffer_last_undop (bp));
         }
       else
-        ms = bp->filename;
-
-      if (write_to_disk (bp, ms))
-        {
-          minibuf_write ("Wrote %s", ms);
-          bp->flags &= ~BFLAG_MODIFIED;
-
-          undo_set_unchanged (bp->last_undop);
-        }
-
-      bp->flags &= ~BFLAG_TEMPORARY;
-
-      if (ms_is_from_minibuffer)
-        free (ms);
+        ok = leNIL;
     }
 
+  if (name_from_minibuffer)
+    free ((char *) name);
+
+  return ok;
+}
+
+static bool
+save_buffer (Buffer * bp)
+{
+  if (!get_buffer_modified (bp))
+    minibuf_write ("(No changes need to be saved)");
+  else
+    write_buffer (bp, get_buffer_needname (bp), false, get_buffer_filename (bp),
+                  "File to save in: ");
+
   return true;
 }
 
@@ -1068,47 +923,15 @@ END_DEFUN
 
 DEFUN ("write-file", write_file)
 /*+
-Write current buffer into the user specified file.
-Makes buffer visit that file, and marks it not modified.
+Write current buffer into file @i{filename}.
+This makes the buffer visit that file, and marks it as not modified.
+
+Interactively, confirmation is required unless you supply a prefix argument.
 +*/
 {
-  char *fname = cur_bp->filename != NULL ? cur_bp->filename : cur_bp->name;
-  char *ms = minibuf_read_filename ("Write file: ", fname, NULL);
-  int ans = true;
-
-  if (ms == NULL)
-    return FUNCALL (keyboard_quit);
-  if (ms[0] == '\0')
-    {
-      free (ms);
-      return leNIL;
-    }
-
-  if (exist_file (ms))
-    {
-      ans = minibuf_read_yn ("File `%s' exists; overwrite? (y or n) ", ms);
-      if (ans == -1)
-        FUNCALL (keyboard_quit);
-      if (ans == false)
-        minibuf_error ("Canceled");
-      if (ans != true)
-        ok = leNIL;
-    }
-
-  if (ans == true)
-    {
-      set_buffer_filename (cur_bp, ms);
-      cur_bp->flags &= ~(BFLAG_NEEDNAME | BFLAG_TEMPORARY);
-      if (write_to_disk (cur_bp, ms))
-        {
-          minibuf_write ("Wrote %s", ms);
-          cur_bp->flags &= ~BFLAG_MODIFIED;
-        }
-      else
-        ok = leNIL;
-    }
-
-  free (ms);
+  ok = write_buffer (cur_bp, true,
+                     arglist && !(lastflag & FLAG_SET_UNIARG),
+                     NULL, "Write file: ");
 }
 END_DEFUN
 
@@ -1116,70 +939,61 @@ static int
 save_some_buffers (void)
 {
   Buffer *bp;
-  int i = 0, noask = false, c;
+  size_t i = 0;
+  bool noask = false;
 
-  for (bp = head_bp; bp != NULL; bp = bp->next)
-    if (bp->flags & BFLAG_MODIFIED && !(bp->flags & BFLAG_NOSAVE))
-      {
-        char *fname = bp->filename != NULL ? bp->filename : bp->name;
+  for (bp = head_bp; bp != NULL; bp = get_buffer_next (bp))
+    {
+      if (get_buffer_modified (bp) && !get_buffer_nosave (bp))
+        {
+          const char *fname = get_buffer_filename_or_name (bp);
 
-        ++i;
+          ++i;
 
-        if (noask)
-          save_buffer (bp);
-        else
-          {
+          if (noask)
+            save_buffer (bp);
+          else
             for (;;)
               {
-                minibuf_write ("Save file %s? (y, n, !, ., q) ", fname);
+                int c;
 
+                minibuf_write ("Save file %s? (y, n, !, ., q) ", fname);
                 c = getkey ();
+                minibuf_clear ();
+
                 switch (c)
                   {
-                  case KBD_CANCEL:
-                  case KBD_RET:
+                  case KBD_CANCEL:	/* C-g */
+                    FUNCALL (keyboard_quit);
+                    return false;
+                  case 'q':
+                    goto endoffunc;
+                  case '.':
+                    save_buffer (bp);
+                    ++i;
+                    return true;
+                  case '!':
+                    noask = true;
+                    /* FALLTHROUGH */
                   case ' ':
                   case 'y':
+                    save_buffer (bp);
+                    ++i;
+                    /* FALLTHROUGH */
                   case 'n':
-                  case 'q':
-                  case '.':
-                  case '!':
+                  case KBD_RET:
+                  case KBD_DEL:
                     goto exitloop;
+                  default:
+                    minibuf_error ("Please answer y, n, !, . or q.");
+                    waitkey (WAITKEY_DEFAULT);
+                    break;
                   }
-
-                minibuf_error ("Please answer y, n, !, . or q.");
-                waitkey (WAITKEY_DEFAULT);
               }
-
-          exitloop:
-            minibuf_clear ();
-
-            switch (c)
-              {
-              case KBD_CANCEL:	/* C-g */
-                FUNCALL (keyboard_quit);
-                return false;
-              case 'q':
-                goto endoffunc;
-              case '.':
-                save_buffer (bp);
-                ++i;
-                return true;
-              case '!':
-                noask = true;
-                /* FALLTHROUGH */
-              case ' ':
-              case 'y':
-                save_buffer (bp);
-                ++i;
-                break;
-              case 'n':
-              case KBD_RET:
-              case KBD_DEL:
-                break;
-              }
-          }
-      }
+        }
+    exitloop:
+      (void)0; /* Label not allowed at end of compound statement. */
+    }
 
 endoffunc:
   if (i == 0)
@@ -1204,25 +1018,24 @@ Offer to save each buffer, then kill this Zile process.
 +*/
 {
   Buffer *bp;
-  int i = 0;
 
   if (!save_some_buffers ())
     return leNIL;
 
-  for (bp = head_bp; bp != NULL; bp = bp->next)
-    if (bp->flags & BFLAG_MODIFIED && !(bp->flags & BFLAG_NEEDNAME))
-      ++i;
-
-  if (i > 0)
-    for (;;)
+  for (bp = head_bp; bp != NULL; bp = get_buffer_next (bp))
+    if (get_buffer_modified (bp) && !get_buffer_needname (bp))
       {
-        int ans = minibuf_read_yesno
-          ("Modified buffers exist; exit anyway? (yes or no) ");
-        if (ans == -1)
-          return FUNCALL (keyboard_quit);
-        else if (!ans)
-          return leNIL;
-        break;
+        for (;;)
+          {
+            int ans = minibuf_read_yesno
+              ("Modified buffers exist; exit anyway? (yes or no) ");
+            if (ans == -1)
+              return FUNCALL (keyboard_quit);
+            else if (!ans)
+              return leNIL;
+            break;
+          }
+        break; /* We have found a modified buffer, so stop. */
       }
 
   thisflag |= FLAG_QUIT;
@@ -1241,16 +1054,14 @@ zile_exit (int doabort)
   Buffer *bp;
 
   fprintf (stderr, "Trying to save modified buffers (if any)...\r\n");
-  for (bp = head_bp; bp != NULL; bp = bp->next)
-    if (bp->flags & BFLAG_MODIFIED && !(bp->flags & BFLAG_NOSAVE))
+  for (bp = head_bp; bp != NULL; bp = get_buffer_next (bp))
+    if (get_buffer_modified (bp) && !get_buffer_nosave (bp))
       {
         astr buf = astr_new (), as;
-        if (bp->filename != NULL)
-          astr_cpy_cstr (buf, bp->filename);
-        else
-          astr_cpy_cstr (buf, bp->name);
+        const char *fname = get_buffer_filename_or_name (bp);
+        astr_cpy_cstr (buf, fname);
         as = astr_new_cstr (PACKAGE);
-        astr_recase (as, UPPERCASE);
+        astr_recase (as, case_upper);
         astr_afmt (buf, ".%sSAVE", astr_cstr (as));
         astr_delete (as);
         fprintf (stderr, "Saving %s...\r\n", astr_cstr (buf));
