@@ -194,6 +194,8 @@ Such a \"function\" cannot be called from Lisp, but it is a valid editor command
 
   /* Copy the keystrokes from cur_mp. */
   append_key_list (mp, cur_mp);
+
+  free(ms);
 }
 END_DEFUN
 
@@ -230,14 +232,14 @@ defining others, use @kbd{M-x name-last-kbd-macro}.
       return leNIL;
     }
 
-  undo_save (UNDO_START_SEQUENCE, cur_bp->pt, 0, 0);
+  undo_save (UNDO_START_SEQUENCE, get_buffer_pt (cur_bp), 0, 0);
   for (uni = 0; uni < uniarg; ++uni)
     call_macro (cur_mp);
-  undo_save (UNDO_END_SEQUENCE, cur_bp->pt, 0, 0);
+  undo_save (UNDO_END_SEQUENCE, get_buffer_pt (cur_bp), 0, 0);
 }
 END_DEFUN
 
-DEFUN_HIDDEN_ARGS ("execute-kbd-macro", execute_kbd_macro,
+DEFUN_NONINTERACTIVE_ARGS ("execute-kbd-macro", execute_kbd_macro,
                    STR_ARG (keystr))
 /*+
 Execute macro as string of editor command characters.
