@@ -30,6 +30,23 @@
 #include "main.h"
 #include "extern.h"
 
+/*
+ * Structure
+ */
+struct Window
+{
+#define FIELD(ty, name) ty name;
+#include "window.h"
+#undef FIELD
+};
+
+#define FIELD(ty, field)                        \
+  GETTER (Window, window, ty, field)            \
+  SETTER (Window, window, ty, field)
+
+#include "window.h"
+#undef FIELD
+
 static Window *
 window_new (void)
 {
@@ -85,7 +102,7 @@ set_current_window (Window * wp)
      marker.  */
   if (cur_wp->saved_pt)
     {
-      cur_bp->pt = cur_wp->saved_pt->pt;
+      set_buffer_pt (cur_bp, get_marker_pt (cur_wp->saved_pt));
       free_marker (cur_wp->saved_pt);
       cur_wp->saved_pt = NULL;
     }
@@ -280,12 +297,7 @@ void
 create_first_window (void)
 {
   Window *wp;
-  Buffer *bp;
-
-  /* Create the scratch buffer. */
-  bp = create_buffer ("*scratch*");
-  bp->flags |= BFLAG_NOSAVE | BFLAG_NEEDNAME | BFLAG_TEMPORARY;
-  cur_bp = bp;
+  Buffer *bp = create_scratch_buffer ();
 
   wp = window_new ();
   cur_wp = head_wp = wp;
@@ -294,7 +306,7 @@ create_first_window (void)
   wp->fheight = term_height () - 1;
   /* Save space for status line. */
   wp->eheight = wp->fheight - 1;
-  wp->bp = bp;
+  wp->bp = cur_bp = bp;
 }
 
 Window *
@@ -303,7 +315,7 @@ find_window (const char *name)
   Window *wp;
 
   for (wp = head_wp; wp != NULL; wp = wp->next)
-    if (!strcmp (wp->bp->name, name))
+    if (!strcmp (get_buffer_name (wp->bp), name))
       return wp;
 
   return NULL;
@@ -321,13 +333,13 @@ window_pt (Window * wp)
       assert (wp->bp == cur_bp);
       assert (wp->saved_pt == NULL);
       assert (cur_bp);
-      return cur_bp->pt;
+      return get_buffer_pt (cur_bp);
     }
   else
     {
       if (wp->saved_pt != NULL)
-        return wp->saved_pt->pt;
+        return get_marker_pt (wp->saved_pt);
       else
-        return wp->bp->pt;
+        return get_buffer_pt (wp->bp);
     }
 }
